!+
! Module: rgssources_update
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!          
! Description: Updates the format of source lists made by versions of rgssources prior to 4.0.
! 
! $Id: rgssources_update.f90,v 1.2 2000/08/11 13:08:45 pvideler Exp $
!- 
module rgssources_update

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal
  use errorHandling

  use pointing_mod

  use constants
!use temp_ssclib
  use ssc_misc	! for removeTrailingChar0s, ucase

  use rgssources_init
  use rgssources_shared

  implicit none 

  type(SpacecraftAttitudeType) :: localAttitude ! typedef from oal
  integer(int16), parameter :: maxNumEpicFiles = 10
  integer(int16) :: epicBand(maxNumEpicFiles)

  private 
  public convertOldFiles

contains 

  !---------------------------------------------------------------------
  subroutine convertOldFiles(setName, writeObsKwds, writeExpKwds&
  , userInstrumentId, paramInstExpIdStr, instrumentId, instExpIdStr)
  !---------------------------------------------------------------------

    implicit none 

    character(*),   intent(in)           :: setName
    logical,        intent(in), optional :: writeObsKwds, writeExpKwds
    integer(int32), intent(in), optional :: userInstrumentId, instrumentId
    character(*),   intent(in), optional :: paramInstExpIdStr, instExpIdStr

    integer(int32) :: localInstrumentId
    character(255) :: localInstExpIdStr
    type(DataSetT) :: set
    type(TableT) :: tab
    character(255) :: instrumentName
    integer(int16) :: maxFormatNum, numSrcs, indexOfPrime, epicTypeNum,&
      numEpicTypes, i
    type(sourceType), dimension(:), allocatable :: srcData
    character(len=8) ::&
      epicExpIdAttrName,&
      epicContentAttrName,&
      epicBandAttrName,&
      epicFilterAttrName
    logical :: localWithObsKwds, localWithExpKwds

    !-------------------------------------------------------------------
    ! Start:

    if (present(writeObsKwds)) then
      localWithObsKwds = writeObsKwds
    else
      localWithObsKwds = .false.
    end if
    if (present(writeExpKwds)) then
      localWithExpKwds = writeExpKwds
    else
      localWithExpKwds = .false.
    end if

    maxFormatNum = size(tableNames) - 1
    formatNum = checkFormat(setName)

    if (formatNum .ne. maxFormatNum) then
      call message(AppMsg, Verbose, 'Converting older-format source &
      &list to sas-5.1 form.')

      set = dataSet(setName, MODIFY)

      if (present(instrumentId)) then
        localInstrumentId = instrumentId
      else if (present(userInstrumentId)) then
        localInstrumentId = userInstrumentId
      else
        call localGetInstrument(setName, set, localInstrumentId)
      end if
      instrumentName = instId(localInstrumentId)

      if (present(instExpIdStr)) then
        localInstExpIdStr = instExpIdStr
      else if (present(paramInstExpIdStr)) then
        localInstExpIdStr = paramInstExpIdStr
      else
        call localGetInstExpId(setName, set, localInstrumentId&
        , localInstExpIdStr)
      end if

      call CAL_setState(instrument=localInstrumentId)
      tab = table(set, tableNames(formatNum))

!!!!!!! TABLE NAME
      call rename(tab, trim(tableNames(maxFormatNum)))

      select case(int(formatNum))
      case(0)
!!!!!!! HEADER KEYWORDS
        localAttitude%ra   = real64Attribute(tab, 'RA_REF')
        localAttitude%dec  = real64Attribute(tab, 'DEC_REF')
        localAttitude%apos = real64Attribute(tab, 'APOS_REF')

        call writeGenericAttributes(set, instrumentName, localAttitude)
        if (localWithObsKwds) then
          call writeObservationAttributes(set)
        end if
        if (localWithExpKwds) then
          call writeExposureAttributes(set, localInstrumentId&
          , localInstExpIdStr)
        end if

!!!!!!! COLUMNS
        indexOfPrime = int16Attribute(tab, 'PRIMESRC')
        numSrcs = numberOfRows(tab)
        allocate(srcData(numSrcs))
        call initSrcDataStructure(srcData)
        call localFillOutputSrcList(tab, indexOfPrime, srcData, numEpicTypes)
        call release(tab) ! just to close all the open columns.
        tab = table(set, tableNames(maxFormatNum))

        call deleteColumn(tab, 'INDEX')
        call deleteColumn(tab, 'LABEL')
        call deleteColumn(tab, 'IGNORE')
        call deleteColumn(tab, 'RA')
        call deleteColumn(tab, 'DEC')
        call deleteColumn(tab, 'RADIUS')
        call deleteColumn(tab, 'ML_ID_SRC')
        call deleteColumn(tab, 'ID_BAND')
        call deleteColumn(tab, 'RATE')
        call deleteColumn(tab, 'RGS_DISP')
        call deleteColumn(tab, 'RGS_XDSP')
        call deleteColumn(tab, 'FOV_PHI')
        call deleteColumn(tab, 'FOV_R')
        call deleteColumn(tab, 'IPRIME')
        call deleteColumn(tab, 'SRC_SELECT')
        call deleteColumn(tab, 'BACK_SELECT')

        call writeNewTable(tab, srcData)

        call deleteColumn(tab, 'FLAG') ! because there isn't one yet in the disk
        ! file format.

!!!!!!! TABLE KEYWORDS
        do epicTypeNum = 1, numEpicTypes
          write(epicExpIdAttrName,   '(a6, i2)') 'E_EXPR', epicTypeNum
          write(epicContentAttrName, '(a6, i2)') 'E_CONT', epicTypeNum
          write(epicFilterAttrName,  '(a6, i2)') 'E_FILT', epicTypeNum
          call setAttribute(tab, epicExpIdAttrName,   'UNKNOWN', COMMEPICEXPID)
          call setAttribute(tab, epicContentAttrName, 'UNKNOWN', COMMEPICCONTENT)
          call setAttribute(tab, epicFilterAttrName,  'UNKNOWN', COMMEPICFILTER)

          ! write(epicBandAttrName,    '(a6, i2)') 'E_BAND', epicTypeNum
          ! call setAttribute(tab, epicBandAttrName, epicBand(epicTypeNum)& !)
          do i=1, size(epicBandIds)
            write(epicBandAttrName,  '(a2,i1,a3,i2)') 'E_',i,'BND',epicTypeNum
            call setAttribute(tab, epicBandAttrName, epicBandIds(i)&
                 , COMMEPICBANDNUM)
          end do
        end do
        call deleteAttribute(tab, 'RA_REF')
        call deleteAttribute(tab, 'DEC_REF')
        call deleteAttribute(tab, 'APOS_REF')

      case default
      end select

      call release(set)
    end if

  end subroutine convertOldFiles

  !---------------------------------------------------------------------
  function checkFormat(setName) result(formatNumber)
  !---------------------------------------------------------------------
  ! This looks at the data set (assumed to be an rgs source list) and tries
  ! to decide which format it adheres to. At present there are only 2
  ! possibilities, corresponding to pre- and post-sas-5.1 respectively.

    implicit none 

    character(*), intent(in) :: setName
    integer(int16) :: formatNumber
    character(len(tableNames)) :: tableName
    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    tableName = name(table(set, 0))
    if (tableName .eq. tableNames(0)) then
      formatNumber = 0
    else if (tableName .eq.  tableNames(1)) then
      formatNumber = 1
    else
call error('badSrcFileType', 'The source list supplied to parameter &
&--rgsset is not in a recognised format.')
    end if
    call release(set)

  end function checkFormat

  !---------------------------------------------------------------------
  subroutine localGetInstrument(setName, set, instrumentId)
  !---------------------------------------------------------------------

    implicit none 

    character(*),   intent(in)  :: setName
    type(DataSetT), intent(in)  :: set
    integer(int32), intent(out) :: instrumentId

    integer(int32) :: fileNameInstrumentId, headerInstrumentId
    logical :: presentInFileName, presentInHeader!!!, presentInParam
    character(20) :: instrumentName

    !-------------------------------------------------------------------
    ! Start:

    if (hasAttribute(set, 'INSTRUME')) then
      instrumentName = stringAttribute(set, 'INSTRUME')
      instrumentId = instId(instrumentName)
  return
    end if

! Try to get it from the command invocation stored in keyword XPROC0:

    presentInHeader = .false. ! default
    if (hasAttribute(set, 'XPROC0')) then
!print *, 'CONTENT  type =', dataType(set, 'CONTENT')
!      xproc0 = stringAttribute(set, 'CONTENT')
!print *, 'CONTENT value =', trim(xproc0)
!print *, ' XPROC0 type  =', dataType(set, 'XPROC0')
!      xproc0 = stringAttribute(set, 'XPROC0')
!print *, ' XPROC0 value =', trim(xproc0)
      headerInstrumentId = -1 ! dummy value to shut the compiler up.
    end if

! Try to get it from file name:

    presentInFileName = .false. ! default
    if (index(setName, 'R1') .ne. 0&
    .or. index(setName, 'R2') .ne. 0) then
      presentInFileName = .true.
      if (index(setName, 'R1') .ne. 0) then
        fileNameInstrumentId = RGS1
      else if (index(setName, 'R2') .ne. 0) then
        fileNameInstrumentId = RGS2
      end if
    end if

    if (presentInHeader .and. presentInFileName&
    .and. headerInstrumentId .ne. fileNameInstrumentId) then
      call warning('instrumentClash', 'The instrument deduced from the &
      &file name does not agree with the instrument in the file header. &
      &Header type will be used.')
    end if

    if (presentInHeader) then
      instrumentId = headerInstrumentId

    else if (presentInFileName) then
      instrumentId = fileNameInstrumentId

      call warning('uncertainInstrument', 'The instrument has been deduced &
      &from the file name.')

    else
call error('unknownInstrument', 'Cannot deduce the instrument from &
&the old-format file.')
    end if

  end subroutine localGetInstrument

  !---------------------------------------------------------------------
  subroutine localGetInstExpId(setName, set, instrumentId, instExpIdStr)
  !---------------------------------------------------------------------

    implicit none 

    character(*),   intent(in)  :: setName
    type(DataSetT), intent(in)  :: set
    integer(int32), intent(in)  :: instrumentId
    character(*),   intent(out) :: instExpIdStr

    character(10) :: obsIdStr
    character(2)  :: insIdStr
    character(1)  :: expTypeChar
    character(3)  :: expIdIntStr
    integer :: status, obsId, dummyInt
    character(4)  :: exposureIdStr

    !-------------------------------------------------------------------
    ! Start:

    if (hasAttribute(set, 'EXPIDSTR')) then
      exposureIdStr = stringAttribute(set, 'EXPIDSTR')

      if (instrumentId .eq. RGS1) then
        instExpIdStr = 'R1'//exposureIdStr
      else if (instrumentId .eq. RGS2) then
        instExpIdStr = 'R2'//exposureIdStr
      else
call error('bug', '')
      end if

  return
    end if

    ! Try to get it from file name:

    if (len(trim(setName)) .ne. 31) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    if (index(setName, 'P') .ne. 1) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    obsIdStr = setName(2:11)
    insIdStr = setName(12:13)
    expTypeChar = setName(14:14)
    expIdIntStr = setName(15:17)

    call strToInt(obsIdStr, obsId, status)
    if (status .ne. 0) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    if (insIdStr .ne. 'R1' .and. insIdStr .ne. 'R2') then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

!    call strToInteger(expIdIntStr, int=dummyInt, status=status)
call strToInt(expIdIntStr, dummyInt, status)
    if (status .ne. 0) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    instExpIdStr = insIdStr//expTypeChar//expIdIntStr

    call warning('uncertainExposure', 'The exposure ID has been deduced &
    &from the file name.')

  end subroutine localGetInstExpId

  !---------------------------------------------------------------------
  subroutine localGetExpId(setName, set, exposureIdStr)
  !---------------------------------------------------------------------

    implicit none 

    character(*),   intent(in)  :: setName
    type(DataSetT), intent(in)  :: set
    character(4), intent(out) :: exposureIdStr

    character(10) :: obsIdStr
    character(2)  :: insIdStr
    character(1)  :: expTypeChar
    character(3)  :: expIdIntStr
    integer :: status, obsId, dummyInt

    !-------------------------------------------------------------------
    ! Start:

    if (hasAttribute(set, 'EXPIDSTR')) then
      exposureIdStr = stringAttribute(set, 'EXPIDSTR')
  return
    end if

    ! Try to get it from file name:

    if (len(trim(setName)) .ne. 31) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    if (index(setName, 'P') .ne. 1) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    obsIdStr = setName(2:11)
    insIdStr = setName(12:13)
    expTypeChar = setName(14:14)
    expIdIntStr = setName(15:17)

!    call strToInteger(obsIdStr, obsId, status)
call strToInt(obsIdStr, obsId, status)
    if (status .ne. 0) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    if (insIdStr .ne. 'R1' .and. insIdStr .ne. 'R2') then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

!    call strToInteger(expIdIntStr, int=dummyInt, status=status)
call strToInt(expIdIntStr, dummyInt, status)
    if (status .ne. 0) then
call error('notStandardFileName', 'The RGS source list does not have a PPS-compliant &
&file name. It is therefore not possible to extract the exposure ID from &
&the file name.')
    end if

    exposureIdStr = expTypeChar//expIdIntStr

    call warning('uncertainExposure', 'The exposure ID has been deduced &
    &from the file name.')

  end subroutine localGetExpId

  !---------------------------------------------------------------------
  subroutine localFillOutputSrcList(tab, indexOfPrime, srcData, numEpicTypes)
  !---------------------------------------------------------------------
  ! Designed to read the source lists produced by versions of rgssources
  ! prior to 4.0 (released in sas-5.1).

    implicit none 

    type(TableT), intent(in) ::&
      tab
    integer(int16), intent(in) ::&
      indexOfPrime
    type(sourceType), dimension(:), intent(out) ::&
      srcData
    integer(int16), intent(out) ::&
      numEpicTypes

    integer(int16) ::&
      numSrcs, numRows, numOffAxisSources, i
    integer(int32) ::&
      src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      mlIdSrcCol,&
      idBandCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      srcSelectCol,&
      backSelectCol
    integer(int8), dimension(:), pointer ::&
      srcSelectColPtr,&
      backSelectColPtr
    integer(int32), dimension(:), pointer ::&
      indexColPtr,&
      mlIdSrcColPtr,&
      idBandColPtr
    real(single), dimension(:), pointer ::&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr
    real(double), dimension(:), pointer ::&
      raColPtr,&
      decColPtr
    character(20) :: label
    logical :: foundBandNumAlready

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)
    numRows = numberOfRows(tab)
if (numSrcs .lt. numRows) call error('bug1','')

    ! Open the columns:
    indexCol      = column(tab, 'INDEX')
    labelCol      = column(tab, 'LABEL')
    raCol         = column(tab, 'RA')
    decCol        = column(tab, 'DEC')
    mlIdSrcCol    = column(tab, 'ML_ID_SRC')
    idBandCol     = column(tab, 'ID_BAND')
    rateCol       = column(tab, 'RATE')
    dispCol       = column(tab, 'RGS_DISP')
    xdspCol       = column(tab, 'RGS_XDSP')
    srcSelectCol  = column(tab, 'SRC_SELECT')
    backSelectCol = column(tab, 'BACK_SELECT')

    ! Open column data values:
    indexColPtr      => int32Data(    indexCol)
    raColPtr         => real64Data(      raCol)
    decColPtr        => real64Data(     decCol)
    mlIdSrcColPtr    => int32Data(  mlIdSrcCol)
    idBandColPtr     => int32Data(   idBandCol)
    rateColPtr       => real32Data(    rateCol)
    dispColPtr       => real32Data(    dispCol)
    xdspColPtr       => real32Data(    xdspCol)
    srcSelectColPtr  => int8Data( srcSelectCol)
    backSelectColPtr => int8Data(backSelectCol)

    numOffAxisSources = 0
    numEpicTypes = 0
    do src = 1, numSrcs
      srcData(src)%index = indexColPtr(src - 1)

      label = trim(removeTrailingChar0s(stringCell(labelCol, src - 1)))
!call stringDump(label)
!call stringDump(removeTrailingChar0s(label))
!call stringDump(trim(removeTrailingChar0s(label)))
!call stringDump('proposal')
      select case(trim(label))
      case('proposal')
        srcData(src)%label  = 'PROPOSAL'!******proposalLabel
        srcData(src)%raDeg  = raColPtr(src - 1)
        srcData(src)%decDeg = decColPtr(src - 1)

      case('onaxis')
        srcData(src)%label = 'ONAXIS'!******onAxisLabel 
        srcData(src)%raDeg  = localAttitude%ra  / piOn180
        srcData(src)%decDeg = localAttitude%dec / piOn180
        srcData(src)%deltaDispArcmin = 0.0
        srcData(src)%deltaXdspArcmin = 0.0
        call dispXdspToPolar(&
          srcData(src)%deltaDispArcmin,&
          srcData(src)%deltaXdspArcmin,&
          srcData(src)%fovPhiDeg,&
          srcData(src)%fovRdeg&
        )
        srcData(src)%fixedOnSky = .false.

      case('epic')
        write(srcData(src)%label, '(a7, i2.2, i5.5)') 'OLDEPIC'&
        , idBandColPtr(src - 1), mlIdSrcColPtr(src - 1)

        srcData(src)%raDeg  = raColPtr(src - 1)
        srcData(src)%decDeg = decColPtr(src - 1)

        foundBandNumAlready = .false. ! default
        do i = 1, numEpicTypes
!!==Change!!!
          if (epicBand(i) .eq. idBandColPtr(src - 1)) then
            foundBandNumAlready = .true.
            srcData(src)%epicFileAttrNum = i
        exit
          end if
        end do
        if (.not. foundBandNumAlready) then
          numEpicTypes = numEpicTypes + 1
          epicBand(numEpicTypes) = idBandColPtr(src - 1)
          srcData(src)%epicFileAttrNum = numEpicTypes
        end if

        !! No check is made any more (since Ver.6.0)
        ! if (idBandColPtr(src - 1) .ne. epicBandNum) then
        !   srcData(src)%flag = iand(srcData(src)%flag, BADBAND_MASK)
        ! end if

      case('offaxis')
        numOffAxisSources = numOffAxisSources + 1

        write(srcData(src)%label, '(a7, i5.5)') 'OFFAXIS'&
        , numOffAxisSources

        srcData(src)%deltaDispArcmin = dispColPtr(src - 1)
        srcData(src)%deltaXdspArcmin = xdspColPtr(src - 1)
        call dispXdspToPolar(&
          srcData(src)%deltaDispArcmin,&
          srcData(src)%deltaXdspArcmin,&
          srcData(src)%fovPhiDeg,&
          srcData(src)%fovRdeg&
        )
        srcData(src)%fixedOnSky = .false.

      case default ! assume user sources
        srcData(src)%label  = ucase(stringCell(labelCol, src - 1))
        srcData(src)%raDeg  = raColPtr(src - 1)
        srcData(src)%decDeg = decColPtr(src - 1)

      end select
      srcData(src)%rate = rateColPtr(src - 1)
      if (srcSelectColPtr(src - 1) .ne. 0) then
        srcData(src)%process = .true.
      end if
      if (backSelectColPtr(src - 1) .ne. 0) then
        srcData(src)%bkgExclude = .true.
      end if
    end do

!*** Note that FOV coords are always recalculated for FOS sources and sky coords always recalc for all others (ONAXIS and OFFAXIS*).

    do src = 1, numSrcs
      if (srcData(src)%fixedOnSky) then ! ra and dec should be left
      ! untouched. Recalculate the others:
        if (srcData(src)%deltaDispArcmin .eq. myReal32null&
        .or. srcData(src)%deltaXdspArcmin .eq. myReal32null) then ! yes.
          call calculateFOVcoords(srcData(src), localAttitude)
        end if

      else ! not 'fixed on sky', the disp/xdisp and fov values should not
      ! be touched. Recalculate the others:
        if (srcData(src)%raDeg .eq. myReal32null&
        .or. srcData(src)%decDeg .eq. myReal32null) then ! yes.
          call invertFOVcoords(srcData(src), localAttitude)
        end if
      end if
    end do

    call calcRgsPlaneConfusion(srcData, indexOfPrime)

  end subroutine localFillOutputSrcList
end module rgssources_update






