#!/usr/local/bin/perl

## Author: Masaaki Sakano (LUX)
## Description:
##   Check the RATE of sources in the EPIC and RGS catalogues
##   for consistency.  If the check fails, exit(2).

use strict;
use Data::Dumper;

## Global variables
our $DEFAULTOUTPUTDIR = "outfiles";

my(@epicdata,@rgsdata);
my($exitstatus) = 0;
my($srcid,$idinst,$idband);
my($line, $band);

my(@WEIGHT)=(0,3.5,1,1);	# (nil, pn, MOS1, MOS2)
my($RGSSTARTID)=3;
my($DEG2RADIAN)=0.0174532925199433;
my($MAXANGLESEPARATION2)=7.71604938271605e-10;	# square of 0.1 arcsec [deg]
my($MAXRATEDIFFRATIO)=0.01;	# [%]
my($VERBOSE)=0;
$VERBOSE=1 if (grep(/^--?v(erbose)?/, @ARGV));

print STDERR "=== Start running rgssources_testrate.pl ===\n" if ($VERBOSE);


## Read EPIC FITS file

open(FH1, "flist $DEFAULTOUTPUTDIR/P0102640401EPX000OMSRLI0000.FIT STDOUT ML_ID_SRC,ID_INST,ID_BAND,RATE,RA,DEC - prhead=no more=yes|") or $exitstatus = 1;
$srcid=$idinst=$idband="nil";	# undefined
@epicdata=();
	# [ML_ID_SRC][ID_INST(non-zero)][ID_BAND] = RATE
	# [ML_ID_SRC][0] = (RA=>.., DEC=>..)
while ($line=<FH1>) {
  chomp $line;
  if ($line =~ /^Row\s*/) {
    $srcid=$idinst=$idband="nil";	# initialise to undefined

  } elsif ($line =~ /^\s+ML_ID_SRC\s*=\s*(\d+)/) {
    $srcid=$1;

  } elsif ($line =~ /^\s+ID_INST\s*=\s*(\d+)/) {
    $idinst=$1;

  } elsif ($line =~ /^\s+ID_BAND\s*=\s*(\d+)/) {
    next if ($idinst == 0);
    $idband=$1;

  } elsif ($line =~ /^\s+RATE\s*=\s*(INDEF|([+\-]?((0|[1-9]\d*)?\.[0-9]+|(0|[1-9]\d*)\.?)([Ee][+\-]?\d+)?))/) {
    next if ($idinst == 0);
    if ($1 eq "INDEF") {
      $epicdata[$srcid][$idinst][$idband] = "nil";
      # printf STDERR "NOTE: No EPIC RATE is defined for the source ML_ID_SRC(%d), Instrument(%d), Band(%d).\n", $srcid, $idinst, $idband if ($VERBOSE);
    } else {
      $epicdata[$srcid][$idinst][$idband] = $2;
    }

  } elsif ($line =~ /^\s+RA\s*=\s*([+\-]?((0|[1-9]\d*)?\.[0-9]+|(0|[1-9]\d*)\.?)([Ee][+\-]?\d+)?)/) {
    next if ($idinst != 0);
    $epicdata[$srcid][0]{"RA"} = $1;

  } elsif ($line =~ /^\s+DEC\s*=\s*([+\-]?((0|[1-9]\d*)?\.[0-9]+|(0|[1-9]\d*)\.?)([Ee][+\-]?\d+)?)/) {
    next if ($idinst != 0);
    $epicdata[$srcid][0]{"DEC"} = $1;
  }
}
close FH1;
#print Dumper \@epicdata;


## Read RGS source catalogue FITS file

my($row);
# open(FH2, "flist ./_0102640401R1S001SRCLI_0000.ds STDOUT LABEL,RA,DEC,RATE - prhead=no more=yes|") or $exitstatus = 1;	# flist causes ERROR, probably due to a bug
open(FH2, "flist ./_0102640401R1S001SRCLI_0000.ds STDOUT RA,DEC,RATE $RGSSTARTID-25 prhead=no more=yes|") or $exitstatus = 1;
$row="nil";	# undefined
@rgsdata=();	# (1 => (RA=>.., DEC=>.., RATE=>..), 2 => (...))
while ($line=<FH2>) {
  chomp $line;
  if ($line =~ /^Row\s*(\d+)/) {
    $row=$1;
    %{$rgsdata[$row]} = ();
  } elsif ($line =~ /^\s+RA\s*=\s*([+\-]?((0|[1-9]\d*)?\.[0-9]+|(0|[1-9]\d*)\.?)([Ee][+\-]?\d+)?)/) {
    $rgsdata[$row]{"RA"} = $1;
  } elsif ($line =~ /^\s+DEC\s*=\s*([+\-]?((0|[1-9]\d*)?\.[0-9]+|(0|[1-9]\d*)\.?)([Ee][+\-]?\d+)?)/) {
    $rgsdata[$row]{"DEC"} = $1;
  } elsif ($line =~ /^\s+RATE\s*=\s*(INDEF|([+\-]?((0|[1-9]\d*)?\.[0-9]+|(0|[1-9]\d*)\.?)([Ee][+\-]?\d+)?))/) {
    if ($1 eq "INDEF") {
      ## Do nothing
    } else {
      $rgsdata[$row]{"RATE"} = $2;
    }
  }
}
close FH2;


## Sum RATEs in the EPIC data for Band 2-3 in each instrument, and write it in
## $epicdata[$srcid][$idinst][0]

my ($rate);
foreach $srcid (1..($#epicdata)) {
  foreach $idinst (1..3) {	# pn, MOS1, MOS2
    $rate="nil";
    foreach $idband (2..3) {	# For Band 2 and 3
      $rate += $epicdata[$srcid][$idinst][$idband] if ($epicdata[$srcid][$idinst][$idband] ne "nil");
    }
    $epicdata[$srcid][$idinst][0] = $rate;

    if ($VERBOSE && (($rate eq "nil") || ($rate eq ""))) {
      printf STDERR "NOTE: No EPIC RATE is defined for the source ML_ID_SRC(%d), Instrument(%d).\n", $srcid, $idinst;
    }
  }
}


## Derive the weighted mean of the RATE from all the EPIC instruments
## appropriately, and compare it with the RGS source one.

my ($epicra, $epicdec, $arinst, $epicrate, $rgsrate);
my ($weightnow, $rowrgsnow);
my ($errmsg);

foreach $srcid (1..($#epicdata)) {

  $epicra  = $epicdata[$srcid][0]{"RA"};
  $epicdec = $epicdata[$srcid][0]{"DEC"};

  ## In fact the following algorithm to calculate the weighted rate
  ## is exactly the same as in rgssources.
  ## In that sense it may be a bit dubious how useful this check is.
  $weightnow=0;
  $epicrate="nil";	# initialise
  foreach $idinst (1..3) {	# pn, MOS1, MOS2
    if ($epicdata[$srcid][$idinst][0] ne "nil") {
      $epicrate += $epicdata[$srcid][$idinst][0];
      $weightnow+=$WEIGHT[$idinst];
    }
  }
  if ($epicrate eq "nil") {
    printf STDERR "WARNING: No EPIC RATE is defined for the source ML_ID_SRC(%d).\n", $srcid;
    next;	# No EPIC RATE is defined.
  } else {
    $epicrate /= $weightnow;
  }

  ## Find a RGS source, of which the position is in agreement with the EPIC one.
  $rgsrate = "nil";
  foreach $row ($RGSSTARTID..($#rgsdata)) {
    if ((($epicra  - $rgsdata[$row]{"RA"})*cos($DEG2RADIAN))**2 + 
        ( $epicdec - $rgsdata[$row]{"DEC"})**2 < $MAXANGLESEPARATION2 ) {
      $rgsrate = $rgsdata[$row]{"RATE"};
      $rowrgsnow = $row;
      last;
    }
  }

  if ($rgsrate eq "nil") {
    printf STDERR "NOTE: No RGS source corresponds to EPIC source ML_ID_SRC(%d).\n", $srcid if ($VERBOSE);
    next;
  } elsif ($rgsrate <= 0) {
    printf STDERR "WARNING: Rate of RGS source (%d) which corresponds to EPIC source ML_ID_SRC(%d) is zero.  Skip.\n", $rowrgsnow, $srcid;
    next;
  }

  if ( abs($epicrate - $rgsrate) > $rgsrate*$MAXRATEDIFFRATIO ) {
    printf STDERR "RGS Row(%d) has a rate (%f), which differs from the correct EPIC ML_ID_SRC(%d) rate (%f) by %f%%.\n", $rowrgsnow, $rgsrate, $srcid, $epicrate, abs($epicrate - $rgsrate)/$rgsrate*100;
    $exitstatus = 2;
  } elsif ($VERBOSE) {
    printf STDERR "RGS Row(%d) has a rate (%f), which differs from the correct EPIC ML_ID_SRC(%d) rate (%f) by %f%% --- OK!\n", $rowrgsnow, $rgsrate, $srcid, $epicrate, abs($epicrate - $rgsrate)/$rgsrate*100;
  }

}


### Manual checking

# my (@tmprate) = (undef, undef, undef, 0.35064331, 0.0196866801818182); #Band2
my (@tmprate) = (undef, undef, undef, 1.21471736, 0.0240079303454545); #Band2+3
my ($i);

## EPIC00001(RGS-Row=3) & EPIC00005(RGS-Row=4)   (See below for Data)
foreach $i (3, 4) {
  if (abs($rgsdata[$i]{"RATE"}-$tmprate[$i])/$tmprate[$i] > 0.01) {
    printf STDERR "ERROR: RGS Row=(%d) has a rate (%f), which should be ~%f.\n", $i, $rgsdata[$i]{"RATE"}, $tmprate[$i];
    $exitstatus = 2;
  }
}


exit($exitstatus);


#################### Data 

### EPIC ($DEFAULTOUTPUTDIR/P0102640401EPX000OMSRLI0000.FIT)
# ML_ID INST BAND RATE          RA                    DEC
#  1     0    0                 2.346182388069586E+01 3.066061046029917E+01
#  1     1    2   INDEF
#  1     1    3   INDEF
## 1     1   2+3  INDEF
#  1     2    2   3.4324664E-01
#  1     2    3   8.0918437E-01
## 1     2   2+3  1.15243101
#  1     3    2   3.5803998E-01
#  1     3    3   9.1896373E-01
## 1     3   2+3  1.27700371
## 1    Avg   2   0.35064331
## 1    Avg  2+3  1.21471736
#
#  5     0    0                 2.329875153003522E+01 3.064532529480380E+01
#  5     1    2   7.3455743E-02
#  5     1    3   1.4180795E-02
## 5     1   2+3  0.087636538
#  5     2    2   1.6124962E-02
#  5     2    3   4.3135621E-03
## 5     2   2+3  0.0204385241
#  5     3    2   1.8696036E-02
#  5     3    3   5.2725188E-03
## 5     3   2+3  0.0239685548
## 5    Avg   2   0.0196866801818182
## 5    Avg  2+3  0.0240079303454545

### RGS (./_0102640401R1S001SRCLI_0000.ds which is created by make test)
# Row LABEL                     RA                    DEC
# 3   EPIC00001                 2.3461824E+01         3.0660610E+01
# 4   EPIC00005                 2.3298752E+01         3.0645325E+01
